"""
ZFSPath.py - main class

Version:      1.0.1
Author:       Danny W. Adair
Last updated: May 06, 2004
License:      ZPL
"""

# ZPT for adding new ZFSPath instances
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

# Base classes of ZFSPath
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager

# Registering a permission
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass

# Helper classes to provide acquisition
from Acquisition import Implicit
from ExtensionClass import Base

# The original path class from path.py
from path import path

# Makes paths to files downloadable
from OFS.Image import File

# Used for conversion of path results
from types import *
PathType = type(path())


manage_addZFSPathForm = PageTemplateFile('zpt/manage_addZFSPathForm', globals())

USE_PERM = 'Use ZFSPaths'


class PathWrapper(Implicit):
    """Wrapper class for path attributes/methods"""
    def __init__(self, pathobject, attr_name):
	self.pathobject = pathobject
	self.attr_name = attr_name

    def _zfspathify(self, result):
	"""
	Converts 'path' instances to ZFSPath instances.
	Handles lists, tuples and dictionaries recursively.
	
	(I.e., listdir() of path.py returns a list of path
	objects. We return a list of ZFSPath objects instead)
	"""
        result_type = type(result)
	if result_type==PathType:
	    return ZFSPath(basepath=str(result)).__of__(self)
	elif result_type in [ListType, TupleType]:
	    if result_type==TupleType:
		result = list(result)
	    new_result = []
	    for element in result:
	        new_result.append(self._zfspathify(element))
	    if result_type==TupleType:
		new_result = tuple(new_result)
	elif result_type==DictType:
	    new_result = {}
	    for key in result.keys():
		new_result[key] = self._zfspathify(result[key])
	else:
	    new_result = result
	return new_result

    def __repr__(self):
	"""
	Redirects representation to attribute access of the contained path object.
	(i.e. "myinstance.name")
	"""
        result = getattr(self.pathobject, self.attr_name)
        return self._zfspathify(result)

    def __call__(self, *args, **kw):
	"""
	Redirects calls (with or without parameters) to the contained path object.
	
	Note:
	PathWrapper instances always appear callable but the underlying
        path instance attribute may not be. We handle this case here.
	"""
	result = getattr(self.pathobject, self.attr_name)
	if callable(result):
	    # If we don't check this then calls like "myinstance/name" in a ZPT fail...
	    # ('__str__ object not callable')
	    result = result(*args, **kw)
	return self._zfspathify(result)

    
def manage_addZFSPath(dispatcher, id, title='', basepath='', REQUEST=None):
    """Adds a new ZFSPath instance"""
    ZFSPath_instance = ZFSPath(id, title, basepath)
    dispatcher._setObject(id, ZFSPath_instance)
    ZFSPath_instance = getattr(dispatcher, id)
    if REQUEST is None:
	# programmatic adding returns the new instance and doesn't redirect
	return ZFSPath_instance
    else:
	# through-the-web adding returns to the ZMI screen of the parent
        return REQUEST.RESPONSE.redirect(dispatcher.absolute_url() + '/manage_main')

    
class ZFSPath(SimpleItem, PropertyManager):
    """ZFSPath class"""
    meta_type = 'ZFSPath'
    _getting_attr = False

    # Connect all attributes to one permission
    security = ClassSecurityInfo()
    security.declareObjectProtected(USE_PERM)

    manage_options = PropertyManager.manage_options + (
                       {'label': 'Documentation',
	                'action':'manage_documentation',
		       },
		     ) + SimpleItem.manage_options

    _properties = (
                   {'id': 'title',    'type':'string', 'mode':'w'},
                   {'id': 'basepath', 'type':'string', 'mode':'w'}
		   )

    manage_documentation = PageTemplateFile('zpt/manage_documentation', globals())

    def __init__(self, id='', title='', basepath=''):
        """Initializes the ZFSPath object."""
        self.id = id
        self.title = title
	self.basepath = basepath

    def __repr__(self):
	"""Simple representation through a 'path' instance"""
        return path(self.basepath).__repr__()

    def __getattr__(self, name):
    	"""
	Returns
	1. path attributes/methods as a 'PathWrapper' instance
	2. files/links to files as a 'File' instance
	3. directories/links to directories as a 'ZFSPath' instance
	4. everything else from the base classes (SimpleItem,
	   PropertyManager), thus allowing for acquisition
	"""
	if name[0]!='_' and not self._getting_attr:
            try:
    		self._getting_attr = True
		
		# path attributes/methods
		pathobject = path(self.basepath)
  	        if hasattr(pathobject, name):
		    return PathWrapper(pathobject, name).__of__(self)

		# files/links/subdirectories
	        pathobject = pathobject.joinpath(name)
	        if pathobject.exists():
		    if pathobject.islink():
			# resolve link
			pathobject = pathobject.readlink()
		    if pathobject.isfile():
			# read file contents and create File object on-the-fly
			# (thus providing resumed downloads and mime type guessing)
                        fs_file = pathobject.open('rb')
			fs_file_content = None
                        try:
			    fs_file_content = fs_file.read()
		        finally:
			    fs_file.close()
			return File(id=pathobject.name, title='', file=fs_file_content).__of__(self)
		    else:
			# return ZFSPath instance 'in the context of me' to allow
			# for conventient recursive access
	                return ZFSPath(basepath = str(pathobject)).__of__(self)
	    finally:
                self._getting_attr = False
	# fall back to base classes
        return Base.__getattr__(self, name)

    def __getitem__(self, name):
	"""
	Provides dictionary-style access to contained files/directories/links.
	Items will always be returned as a ZFSPath instance.
	"""
	pathobject = path(self.basepath).joinpath(name)
	if pathobject.exists():
	    return ZFSPath(basepath = str(pathobject)).__of__(self)
	else:
	    raise KeyError, name

InitializeClass(ZFSPath)
